# Type-Hinting is done in a stub file
"""
Overview
========

This module contains the encoding/decoding logic for data types as defined in
:term:`X.690`.

Each type is made available via a registry dictionary on :py:class:`~.Type` and
can be retrieved via :py:meth:`~.Type.get`.

Additionally, given a :py:class:`bytes` object, the :py:func:`~.decode`
function can be used to parse the bytes object and return a typed instance
from it. See :py:func:`~.decode` for details about it's behaviour!

.. note::
    The individual type classes in this module do not contain any additional
    documentation. The bulk of this module is documented in :py:class:`~.Type`.

    For the rest, the type classes simply define the type identifier tag.

Supporting Additional Classes
=============================

Just by subclassing :py:class:`~.Type` and setting correct ``TAG`` and
``TYPECLASS`` values, most of the basic functionality will be covered by the
superclass. Type detection, and addition to the registry is automatic.
Subclassing is enough.

By default, a new type which does not override any methods will have it's value
reported as bytes objects. You may want to override at least
:py:meth:`~.Type.decode_raw` to convert the raw-bytes into your own data-type.

Example
-------

Let's assume you want to decode/encode a "Person" object with a first-name,
last-name and age. Let's also assume it will be an application-specific type of
a "constructed" nature with our application-local tag 1. Let's further assume
that the value will be a UTF-8 encoded JSON string inside the x690 stream.

We specify the metadata as class-level variables ``TYPECLASS``, ``NATURE`` and
``TAG``. The decoding is handled by implementing a static-method
``decode_raw`` which gets the data-object containing the value and a slice
defining at which position the data is located. The encoding is handled by
implementing the instance-method ``encode_raw``. The instance contains the
Python value in ``self.pyvalue``.

So we can implement this as follows (including a named-tuple as our local
type):

.. code-block:: python

    from typing import NamedTuple
    from x690.types import Type
    from json import loads, dumps

    class Person(NamedTuple):
        first_name: str
        last_name: str
        age: int

    class PersonType(Type[Person]):
        TYPECLASS = TypeClass.APPLICATION
        NATURE = [TypeNature.CONSTRUCTED]
        TAG = 1

        @staticmethod
        def decode_raw(data: bytes, slc: slice = slice(None)) -> Person:
            values = loads(data[slc].decode("utf8"))
            return Person(
                values["first_name"], values["last_name"], values["age"]
            )

        def encode_raw(self) -> bytes:
            return dumps(self.pyvalue._asdict()).encode("utf8")


"""
# pylint: disable=abstract-method, missing-class-docstring, too-few-public-methods


from datetime import datetime
from itertools import zip_longest
from textwrap import indent
from typing import Any, Dict, Generic, Iterator, List, Optional, Tuple
from typing import Type as TypeType
from typing import TypeVar, Union

import t61codec  # type: ignore

from .exc import IncompleteDecoding, UnexpectedType
from .util import (
    INDENT_STRING,
    TypeClass,
    TypeInfo,
    TypeNature,
    encode_length,
    get_value_slice,
    visible_octets,
    wrap,
)

TWrappedPyType = TypeVar("TWrappedPyType", bound=Any)
TPopType = TypeVar("TPopType", bound=Any)


def decode(
    data: bytes,
    start_index: int = 0,
    enforce_type: Optional[TypeType[TPopType]] = None,
    strict: bool = False,
) -> Tuple[Union["TPopType", "Null"], int]:
    """
    Convert a X.690 bytes object into a Python instance, and the location of
    the next object.

    Given a :py:class:`bytes` object and any start-index, inspects and parses
    the octets starting at the given index (as many as required) to determine
    variable type (and corresponding Python class), and length. That class is
    then used to parse the object located in ``data`` at the given index. The
    location of the start of the next (subsequent) object is also determined.

    The return value is a tuple with the decoded object and the start-index
    of the next object.

    Example::

        >>> data = b'\\x02\\x01\\x05\\x11'
        >>> decode(data)
        (Integer(5), 3)
        >>> data = b'some-skippable-bytes\\x02\\x01\\x05\\x11'
        >>> decode(data, 20)
        (Integer(5), 23)
    """
    if start_index >= len(data):
        return Null(), 0

    start_index = start_index or 0
    type_ = TypeInfo.from_bytes(data[start_index])

    try:
        cls = Type.get(type_.cls, type_.tag, type_.nature)
    except KeyError:
        cls = UnknownType

    data_slice, next_tlv = get_value_slice(data, start_index)
    output = cls.from_bytes(data, data_slice)
    if cls is UnknownType:
        output.tag = data[start_index]  # type: ignore

    if enforce_type and not isinstance(output, enforce_type):
        raise UnexpectedType(
            f"Unexpected decode result. Expected instance of type "
            f"{enforce_type} but got {type(output)} instead"
        )

    if strict and next_tlv < len(data) - 1:
        remainder = data[next_tlv:]
        raise IncompleteDecoding(
            f"Strict decoding still had {len(remainder)} remaining bytes!",
            remainder=remainder,
        )

    return output, next_tlv  # type: ignore


class Type(Generic[TWrappedPyType]):
    """
    The superclass for all supported types.
    """

    __slots__ = ["pyvalue"]
    __registry: Dict[Tuple[str, int, TypeNature], TypeType["Type[Any]"]] = {}

    #: The x690 type-class (universal, application or context)
    TYPECLASS: TypeClass = TypeClass.UNIVERSAL

    #: The x690 "private/constructed" information
    NATURE = [TypeNature.CONSTRUCTED]

    #: The x690 identifier for the type
    TAG: int = -1

    #: The decoded (or to-be encoded) Python value
    pyvalue: Optional[TWrappedPyType]

    #: The byte representation of "pyvalue" without metadata-header
    raw_bytes: bytes

    #: The location of the value within "raw_bytes"
    bounds: slice = slice(None)

    def __init_subclass__(cls: TypeType["Type[Any]"]) -> None:
        if cls.__name__ == "Type" and cls.TAG == -1:
            return
        for nature in cls.NATURE:
            Type.__registry[(cls.TYPECLASS, cls.TAG, nature)] = cls

    @property
    def value(self) -> TWrappedPyType:
        """
        Returns the value as a pure Python type
        """
        return self.pyvalue or self.decode_raw(self.raw_bytes, self.bounds)

    @staticmethod
    def decode_raw(data: bytes, slc: slice = slice(None)) -> TWrappedPyType:
        """
        Converts the raw byte-value (without type & length header) into a
        pure Python type

        >>> Integer.decode_raw(b"\\x05")
        5

        :param data: A data-block containing the byte-information
        :param slc: A slice of the data-block that contains the exact
            raw-bytes.
        :return: The value that should be wrapped by the current x690 type.
        """
        return data[slc]  # type: ignore

    @staticmethod
    def get(
        typeclass: str, typeid: int, nature: TypeNature = TypeNature.CONSTRUCTED
    ) -> TypeType["Type[Any]"]:
        """
        Retrieve a Python class by x690 type information

        Classes can be registered by subclassing :py:class:`x690.types.Type`
        """
        cls = Type.__registry[(typeclass, typeid, nature)]
        return cls

    @staticmethod
    def all() -> List[TypeType["Type[Any]"]]:
        """
        Returns all registered classes
        """
        return list(Type.__registry.values())

    @classmethod
    def validate(cls, data: bytes) -> None:
        """
        Given a bytes object, checks if the given class *cls* supports decoding
        this object. If not, raises a ValueError.
        """
        # TODO: Making this function return a boolean instead of raising an
        #       exception would make the code potentially more readable.
        tinfo = TypeInfo.from_bytes(data[0])
        if tinfo.cls != cls.TYPECLASS or tinfo.tag != cls.TAG:
            raise ValueError(
                "Invalid type header! "
                "Expected a %s class with tag "
                "ID 0x%02x, but got a %s class with "
                "tag ID 0x%02x" % (cls.TYPECLASS, cls.TAG, tinfo.cls, data[0])
            )

    @classmethod
    def decode(cls, data: bytes) -> "Type[TWrappedPyType]":  # pragma: no cover
        """
        This method takes a bytes object which contains the raw content octets
        of the object. That means, the octets *without* the type information
        and length.

        This function must be overridden by the concrete subclasses.
        """
        slc = get_value_slice(data).bounds
        output = cls.decode_raw(data, slc)
        return cls(output)

    @classmethod
    def from_bytes(
        cls, data: bytes, slc: slice = slice(None)
    ) -> "Type[TWrappedPyType]":
        """
        Creates a new :py:class:`x690.types.Type` instance from raw-bytes
        (without type nor length bytes)

        >>> Integer.from_bytes(b"\\x01")
        Integer(1)
        >>> OctetString.from_bytes(b"hello-world")
        OctetString(b'hello-world')
        >>> Boolean.from_bytes(b"\\x00")
        Boolean(False)
        """
        instance = cls()
        instance.raw_bytes = data
        instance.bounds = slc
        return instance

    def __init__(self, value: Optional[TWrappedPyType] = None) -> None:
        self.pyvalue = value
        if value is None:
            self.raw_bytes = b""
        else:
            self.raw_bytes = self.encode_raw()

    def __bytes__(self) -> bytes:  # pragma: no cover
        """
        Convert this instance into a bytes object. This must be implemented by
        subclasses.
        """
        value = self.raw_bytes[self.bounds] or self.encode_raw()
        tinfo = TypeInfo(self.TYPECLASS, self.NATURE[0], self.TAG)
        return bytes(tinfo) + encode_length(len(value)) + value

    def __repr__(self) -> str:
        return "%s(%r)" % (self.__class__.__name__, self.value)

    @property
    def length(self) -> int:
        """
        Return the x690 byte-length of this instance
        """
        return len(self.raw_bytes[self.bounds])

    def encode_raw(self) -> bytes:
        """
        Convert this instance into raw x690 bytes (excluding the type and
        length header)

        >>> Integer(5).encode_raw()
        b'\\x05'
        >>> Boolean(True).encode_raw()
        b'\\x01'
        """
        if self.pyvalue is None:
            return b""
        return self.pyvalue

    def pythonize(self) -> TWrappedPyType:
        """
        Convert this instance to an appropriate pure Python object.
        """
        return self.value

    def pretty(self, depth: int = 0) -> str:  # pragma: no cover
        """
        Returns a readable representation (possibly multiline) of the value.

        The value is indented by *depth* levels of indentation

        By default this simply returns the string representation. But more
        complex values may override this.
        """
        return indent(str(self), INDENT_STRING * depth)


class UnknownType(Type[bytes]):
    """
    A fallback type for anything not in X.690.

    Instances of this class contain the raw information as parsed from the
    bytes as the following attributes:

    * ``value``: The value without leading metadata (as bytes value)
    * ``tag``: The *unparsed* "tag". This is the type ID as defined in the
      reference document. See :py:class:`~puresnmp.x690.util.TypeInfo` for
      details.
    * ``typeinfo``: unused (derived from *tag* and only here for consistency
      with ``__repr__`` of this class).
    """

    TAG = 0x99

    def __init__(self, value: bytes = b"", tag: int = -1) -> None:
        super().__init__(value)
        self.tag = tag

    def __repr__(self) -> str:
        typeinfo = TypeInfo.from_bytes(self.tag)
        tinfo = f"{typeinfo.cls}/{typeinfo.nature}/{typeinfo.tag}"
        return f"<{self.__class__.__name__} {self.tag} {self.value!r} {tinfo}>"

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, UnknownType)
            and self.value == other.value
            and self.tag == other.tag
        )

    def pretty(self, depth: int = 0) -> str:
        """
        Returns a prettified string with *depth* levels of indentation

        See :py:meth:`~.Type.pretty`
        """
        wrapped = wrap(
            visible_octets(self.value), str(type(self)), depth
        ).splitlines()
        if len(wrapped) > 15:
            line_width = len(wrapped[0])
            sniptext = ("<%d more lines>" % (len(wrapped) - 10 - 5)).center(
                line_width - 2
            )
            wrapped = wrapped[:10] + ["┊%s┊" % sniptext] + wrapped[-5:]
        typeinfo = TypeInfo.from_bytes(self.tag)
        lines = [
            "Unknown Type",
            f"  │ Tag:       {self.tag}",
            "  │ Type Info:",
            f"  │  │ Class: {typeinfo.cls}",
            f"  │  │ Nature: {typeinfo.nature}",
            f"  │  │ Tag: {typeinfo.tag}",
        ] + wrapped
        return indent(
            "\n".join(lines),
            INDENT_STRING * depth,
        )


class Boolean(Type[bool]):
    TAG = 0x01
    NATURE = [TypeNature.PRIMITIVE]

    @staticmethod
    def decode_raw(data: bytes, slc: slice = slice(None)) -> bool:
        """
        Converts the raw byte-value (without type & length header) into a
        pure Python type

        Overrides :py:meth:`~.Type.decode_raw`
        """
        return data[slc] != b"\x00"

    @classmethod
    def validate(cls, data: bytes) -> None:
        """
        Overrides :py:meth:`.Type.validate`
        """
        super().validate(data)
        if data[1] != 1:
            raise ValueError(
                "Unexpected Boolean value. Length should be 1,"
                " it was %d" % data[1]
            )

    def encode_raw(self) -> bytes:
        """
        Overrides :py:meth:`.Type.encode_raw`
        """
        return b"\x01" if self.pyvalue else b"\x00"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Boolean) and self.value == other.value


class Null(Type[None]):
    TAG = 0x05
    NATURE = [TypeNature.PRIMITIVE]

    @classmethod
    def validate(cls, data: bytes) -> None:
        """
        Overrides :py:meth:`.Type.validate`
        """
        super().validate(data)
        if data[1] != 0:
            raise ValueError(
                "Unexpected NULL value. Length should be 0, it "
                "was %d" % data[1]
            )

    @staticmethod
    def decode_raw(data: bytes, slc: slice = slice(None)) -> None:
        """
        Converts the raw byte-value (without type & length header) into a
        pure Python type

        Overrides :py:meth:`~.Type.decode_raw`
        """
        # pylint: disable=unused-argument
        return None

    def encode_raw(self) -> bytes:
        """
        Overrides :py:meth:`.Type.encode_raw`
        """
        # pylint: disable=no-self-use
        return b"\x00"

    def __bytes__(self) -> bytes:
        return b"\x05\x00"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Null) and self.value == other.value

    def __repr__(self) -> str:
        return "Null()"

    def __bool__(self) -> bool:
        return False

    def __nonzero__(self) -> bool:  # pragma: no cover
        return False


class OctetString(Type[bytes]):
    TAG = 0x04
    NATURE = [TypeNature.PRIMITIVE, TypeNature.CONSTRUCTED]

    def __init__(self, value: Union[str, bytes] = b"") -> None:
        if isinstance(value, str):
            value = value.encode("ascii")
        super().__init__(value)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, OctetString) and self.value == other.value

    def pretty(self, depth: int = 0) -> str:
        """
        Returns a prettified string with *depth* levels of indentation

        See :py:meth:`~.Type.pretty`
        """
        if self.value == b"":
            return repr(self)
        try:
            # We try to decode embedded X.690 items. If we can't, we display
            # the value raw
            embedded = decode(self.value)[0]
            if isinstance(embedded, UnknownType):
                raise TypeError("UnknownType should not be prettified here")
            return wrap(embedded.pretty(0), f"Embedded in {type(self)}", depth)
        except:  # pylint: disable=bare-except
            wrapped = wrap(visible_octets(self.value), str(type(self)), depth)
            return wrapped


class Sequence(Type[List[Type[Any]]]):
    """
    Represents an X.690 sequence type. Instances of this class are iterable and
    indexable.
    """

    TAG = 0x10

    @staticmethod
    def decode_raw(data: bytes, slc: slice = slice(None)) -> List[Type[Any]]:
        """
        Converts the raw byte-value (without type & length header) into a
        pure Python type

        Overrides :py:meth:`~.Type.decode_raw`
        """
        if not data or slc.start > len(data):
            return []
        item, next_pos = decode(data, slc.start)
        items: List[Type[Any]] = [item]
        end = slc.stop or len(data)
        while next_pos < end:
            item, next_pos = decode(data, next_pos)
            items.append(item)
        return items

    def encode_raw(self) -> bytes:
        """
        Overrides :py:meth:`.Type.encode_raw`
        """
        if self.pyvalue is None:
            return b""
        items = [bytes(item) for item in self.pyvalue]
        output = b"".join(items)
        return output

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Sequence):
            return False
        return self.raw_bytes[self.bounds] == other.raw_bytes[other.bounds]

    def __repr__(self) -> str:
        item_repr = list(self)
        return "Sequence(%r)" % item_repr

    def __len__(self) -> int:
        return len(self.value)

    def __iter__(self) -> Iterator[Type[Any]]:
        yield from self.value

    def __getitem__(self, idx: int) -> Type[Any]:
        return self.value[idx]

    def pythonize(self) -> List[Type[Any]]:
        """
        Overrides :py:meth:`~.Type.pythonize`
        """
        return [obj.pythonize() for obj in self]

    def pretty(self, depth: int = 0) -> str:  # pragma: no cover
        """
        Returns a prettified string with *depth* levels of indentation

        See :py:meth:`~.Type.pretty`
        """
        lines = [f"{self.__class__.__name__} with {len(self.value)} items:"]
        for item in self.value:
            prettified_item = item.pretty(depth)
            bullet = INDENT_STRING * depth + "⁃ "
            for line in prettified_item.splitlines():
                lines.append(bullet + line)
                bullet = "  "
        return "\n".join(lines)


class Integer(Type[int]):
    SIGNED = True
    TAG = 0x02
    NATURE = [TypeNature.PRIMITIVE]

    @staticmethod
    def decode_raw(data: bytes, slc: slice = slice(None)) -> int:
        """
        Converts the raw byte-value (without type & length header) into a
        pure Python type

        Overrides :py:meth:`~.Type.decode_raw`
        """
        data = data[slc]
        return int.from_bytes(data, "big", signed=Integer.SIGNED)

    def encode_raw(self) -> bytes:
        """
        Overrides :py:meth:`.Type.encode_raw`
        """
        if self.pyvalue is None:
            return b""
        octets = [self.pyvalue & 0b11111111]

        # Append remaining octets for long integers.
        remainder = self.pyvalue
        while remainder not in (0, -1):
            remainder = remainder >> 8
            octets.append(remainder & 0b11111111)

        if remainder == 0 and octets[-1] == 0b10000000:
            octets.append(0)
        octets.reverse()

        # remove leading octet if there is a string of 9 zeros or ones
        while len(octets) > 1 and (
            (octets[0] == 0 and octets[1] & 0b10000000 == 0)
            or (octets[0] == 0b11111111 and octets[1] & 0b10000000 != 0)
        ):
            del octets[0]
        return bytes(octets)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Integer) and self.value == other.value


class ObjectIdentifier(Type[Tuple[int, ...]]):
    """
    Represents an OID.

    Instances of this class support containment checks to determine if one OID
    is a sub-item of another::

        >>> ObjectIdentifier((1, 2, 3, 4, 5)) in ObjectIdentifier((1, 2, 3))
        True

        >>> ObjectIdentifier((1, 2, 4, 5, 6)) in ObjectIdentifier((1, 2, 3))
        False
    """

    TAG = 0x06
    NATURE = [TypeNature.PRIMITIVE]

    @staticmethod
    def decode_large_value(current_char: int, stream: Iterator[int]) -> int:

        """
        If we encounter a value larger than 127, we have to consume from the
        stram until we encounter a value below 127 and recombine them.

        See: https://msdn.microsoft.com/en-us/library/bb540809(v=vs.85).aspx
        """
        buffer = []
        while current_char > 127:
            buffer.append(current_char ^ 0b10000000)
            current_char = next(stream)
        total = current_char
        for i, digit in enumerate(reversed(buffer)):
            total += digit * 128 ** (i + 1)
        return total

    @staticmethod
    def encode_large_value(value: int) -> List[int]:
        """
        Inverse function of :py:meth:`~.ObjectIdentifier.decode_large_value`
        """
        if value <= 127:
            return [value]
        output = [value & 0b1111111]
        value = value >> 7
        while value:
            output.append(value & 0b1111111 | 0b10000000)
            value = value >> 7
        output.reverse()
        return output

    @staticmethod
    def decode_raw(data: bytes, slc: slice = slice(None)) -> Tuple[int, ...]:
        """
        Converts the raw byte-value (without type & length header) into a
        pure Python type

        Overrides :py:meth:`~.Type.decode_raw`
        """
        # Special case for "empty" object identifiers which should be returned
        # as "0"
        data = data[slc]
        if not data:
            return (0,)

        # unpack the first byte into first and second sub-identifiers.
        data0 = data[0]
        first, second = data0 // 40, data0 % 40
        output = [first, second]

        remaining = iter(data[1:])

        for node in remaining:
            # Each node can only contain values from 0-127. Other values need
            # to be combined.
            if node > 127:
                collapsed_value = ObjectIdentifier.decode_large_value(
                    node, remaining
                )
                output.append(collapsed_value)
                continue
            output.append(node)

        instance = tuple(output)
        return instance

    @staticmethod
    def from_string(value: str) -> "ObjectIdentifier":
        """
        Create an OID from a string
        """

        if value == ".":
            return ObjectIdentifier((1,))

        if value.startswith("."):
            value = value[1:]

        identifiers = tuple(int(ident, 10) for ident in value.split("."))
        return ObjectIdentifier(identifiers)

    def collapse_identifiers(
        self, identifiers: Tuple[int, ...]
    ) -> Tuple[int, ...]:
        """
        Meld the first two octets into one octet as defined by x.690

        In x.690 ObjectIdentifiers are a sequence of numbers. In the
        byte-representation the first two of those numbers are stored in the
        first byte.

        This function takes a "human-readable" OID tuple and returns a new
        tuple with the first two elements merged (collapsed) together.
        """
        # pylint: disable=no-self-use
        if len(identifiers) == 0:
            return tuple()

        if len(identifiers) > 1:
            # The first two bytes are collapsed according to X.690
            # See https://en.wikipedia.org/wiki/X.690#BER_encoding
            first, second, rest = (
                identifiers[0],
                identifiers[1],
                identifiers[2:],
            )
            first_output = (40 * first) + second
        else:
            first_output = identifiers[0]
            rest = tuple()

        # Values above 127 need a special encoding. They get split up into
        # multiple positions.
        exploded_high_values = []
        for char in rest:
            if char > 127:
                exploded_high_values.extend(
                    ObjectIdentifier.encode_large_value(char)
                )
            else:
                exploded_high_values.append(char)

        collapsed_identifiers = [first_output]
        for subidentifier in rest:
            collapsed_identifiers.extend(
                ObjectIdentifier.encode_large_value(subidentifier)
            )
        return tuple(collapsed_identifiers)

    def encode_raw(self) -> bytes:
        """
        Overrides :py:meth:`.Type.encode_raw`
        """
        if self.pyvalue is None:
            return b""
        collapsed_identifiers = self.collapse_identifiers(self.pyvalue)
        if collapsed_identifiers == (0,):
            return b""
        return bytes(collapsed_identifiers)

    def __int__(self) -> int:
        if len(self.value) != 1:
            raise ValueError(
                "Only ObjectIdentifier with one node can be "
                "converted to int. %r is not convertable" % self
            )
        return self.value[0]

    def __str__(self) -> str:
        return ".".join([str(_) for _ in self.value])

    def __repr__(self) -> str:
        return "ObjectIdentifier(%r)" % (self.value,)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ObjectIdentifier) and self.value == other.value

    def __len__(self) -> int:
        return len(self.value)

    def __contains__(self, other: "ObjectIdentifier") -> bool:
        """
        Check if one OID is a child of another.

        TODO: This has been written in the middle of the night! It's messy...
        """
        # pylint: disable=invalid-name

        a, b = other.value, self.value

        # if both have the same amount of identifiers, check for equality
        if len(a) == len(b):
            return a == b

        # if "self" is longer than "other", self cannot be "in" other
        if len(b) > len(a):
            return False

        # For all other cases:
        #   1. zero-fill
        #   2. drop identical items from the front (leaving us with "tail")
        #   3. compare both tails
        zipped = zip_longest(a, b, fillvalue=None)
        tail: List[Tuple[int, int]] = []
        for tmp_a, tmp_b in zipped:
            if tmp_a == tmp_b and not tail:
                continue
            tail.append((tmp_a, tmp_b))

        # if we only have Nones in "b", we know that "a" was longer and that it
        # is a direct subtree of "b" (no diverging nodes). Otherwise we would
        # have te divergence in "b", and we can say that "b is contained in a"
        _, unzipped_b = zip(*tail)
        if all([x is None for x in unzipped_b]):
            return True

        # In all other cases we end up with an unmatching tail and know that "b
        # is not contained in a".
        return False

    def __lt__(self, other: "ObjectIdentifier") -> bool:
        return self.value < other.value

    def __hash__(self) -> int:
        return hash(self.value)

    def __add__(self, other: "ObjectIdentifier") -> "ObjectIdentifier":
        nodes = self.value + other.value
        return ObjectIdentifier(nodes)

    def __getitem__(self, index: int) -> "ObjectIdentifier":
        return ObjectIdentifier((self.value[index],))

    def parentof(self, other: "ObjectIdentifier") -> bool:
        """
        Convenience method to check whether this OID is a parent of another OID
        """
        return other in self

    def childof(self, other: "ObjectIdentifier") -> bool:
        """
        Convenience method to check whether this OID is a child of another OID
        """
        return self in other


class ObjectDescriptor(Type[str]):
    TAG = 0x07
    NATURE = [TypeNature.PRIMITIVE, TypeNature.CONSTRUCTED]


class External(Type[bytes]):
    TAG = 0x08


class Real(Type[float]):
    TAG = 0x09
    NATURE = [TypeNature.PRIMITIVE]


class Enumerated(Type[List[Any]]):
    TAG = 0x0A
    NATURE = [TypeNature.PRIMITIVE]


class EmbeddedPdv(Type[bytes]):
    TAG = 0x0B


class Utf8String(Type[str]):
    TAG = 0x0C
    NATURE = [TypeNature.PRIMITIVE, TypeNature.CONSTRUCTED]


class RelativeOid(Type[str]):
    TAG = 0x0D
    NATURE = [TypeNature.PRIMITIVE]


class Set(Type[bytes]):
    TAG = 0x11


class NumericString(Type[str]):
    TAG = 0x12
    NATURE = [TypeNature.PRIMITIVE, TypeNature.CONSTRUCTED]


class PrintableString(Type[str]):
    TAG = 0x13
    NATURE = [TypeNature.PRIMITIVE, TypeNature.CONSTRUCTED]


class T61String(Type[str]):
    TAG = 0x14
    NATURE = [TypeNature.PRIMITIVE, TypeNature.CONSTRUCTED]
    __INITIALISED = False

    def __init__(self, value: Union[str, bytes] = "") -> None:
        if isinstance(value, str):
            super().__init__(value)
        else:
            super().__init__(T61String.decode_raw(value))

    def __eq__(self, other: object) -> bool:
        return isinstance(other, T61String) and self.value == other.value

    @staticmethod
    def decode_raw(data: bytes, slc: slice = slice(None, None)) -> str:
        """
        Converts the raw byte-value (without type & length header) into a
        pure Python type

        Overrides :py:meth:`~.Type.decode_raw`
        """
        data = data[slc]
        if not T61String.__INITIALISED:
            t61codec.register()
            T61String.__INITIALISED = True
        return data.decode("t61")

    def encode_raw(self) -> bytes:
        """
        Overrides :py:meth:`.Type.encode_raw`
        """
        if not T61String.__INITIALISED:
            t61codec.register()
            T61String.__INITIALISED = True
        if self.pyvalue is None:
            return b""
        return self.pyvalue.encode("t61")


class VideotexString(Type[str]):
    TAG = 0x15
    NATURE = [TypeNature.PRIMITIVE, TypeNature.CONSTRUCTED]


class IA5String(Type[str]):
    TAG = 0x16
    NATURE = [TypeNature.PRIMITIVE, TypeNature.CONSTRUCTED]


class UtcTime(Type[datetime]):
    TAG = 0x17
    NATURE = [TypeNature.PRIMITIVE, TypeNature.CONSTRUCTED]


class GeneralizedTime(Type[datetime]):
    TAG = 0x18
    NATURE = [TypeNature.PRIMITIVE, TypeNature.CONSTRUCTED]


class GraphicString(Type[str]):
    # NOTE: As per x.690, this should inherit from OctetString. However, this
    #       library serves as an abstraction layer between X.690 and Python.
    #       For this reason, it defines this as a "str" type. To keep the
    #       correct behaviours, we can still "borrow" the implementation from
    #       OctetString if needed
    TAG = 0x19
    NATURE = [TypeNature.PRIMITIVE, TypeNature.CONSTRUCTED]

    @staticmethod
    def decode_raw(data: bytes, slc: slice = slice(None)) -> str:
        """
        Converts the raw byte-value (without type & length header) into a
        pure Python type

        Overrides :py:meth:`~.Type.decode_raw`
        """
        data = data[slc]
        return data.decode("ascii")


class VisibleString(Type[str]):
    TAG = 0x1A
    NATURE = [TypeNature.PRIMITIVE, TypeNature.CONSTRUCTED]


class GeneralString(Type[str]):
    TAG = 0x1B
    NATURE = [TypeNature.PRIMITIVE, TypeNature.CONSTRUCTED]


class UniversalString(Type[str]):
    TAG = 0x1C
    NATURE = [TypeNature.PRIMITIVE, TypeNature.CONSTRUCTED]


class CharacterString(Type[str]):
    TAG = 0x1D


class BmpString(Type[str]):
    TAG = 0x1E
    NATURE = [TypeNature.PRIMITIVE, TypeNature.CONSTRUCTED]


class EOC(Type[bytes]):
    TAG = 0x00
    NATURE = [TypeNature.PRIMITIVE]


class BitString(Type[str]):
    TAG = 0x03
    NATURE = [TypeNature.PRIMITIVE, TypeNature.CONSTRUCTED]



"""
x690 is a library allowing to decode and encode values according to the ITU
X.690 standard.

For decoding, use the function *decode* provided by this module. When used
as-is on a bytes object, it will return the first decoded value in the blob and
the position of the next value:

>>> from x690 import decode
>>> data, next_index = decode(b"\x02\01\01\x02\01\03")
>>> data
Integer(1)
>>> next_index
3

For a stream of data with an unknown number of items, you can use the
next-index to move over the data:

>>> from x690 import decode
>>> data = b"\\x02\\x01\\x01\\x02\\x01\\x02\\x02\\x01\\x03"
>>> item, next_index = decode(data)
>>> item, next_index
(Integer(1), 3)
>>> while next_index < len(data):
...     item, next_index = decode(data, next_index)
...     item, next_index
(Integer(2), 6)
(Integer(3), 9)

If you expect exactly one element, it is possible to pass the "strict"
argument. This will raise an error if the stream contains any "junk" data.

Finally, a x690 data-stream may contain various types. For this reason, the
"decode" function is type-hinted to return a vague, imprecise type. If you
*know* what your are decoding, you can pass the expected type into the
function:

>>> from x690 import decode
>>> from x690.types import Integer
>>> decode(b"\x02\x01\x01", enforce_type=Integer)
(Integer(1), 3)

This will do two things:

* It runs a type-check upon decoding and will raise a
  :py:exc:`x690.exc.UnexpectedType` error if it does not match.
* Inform the type-checker of the return-type, improving type-checker output.
"""

from .types import decode

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata  # type: ignore


__version__ = importlib_metadata.version("x690")
__all__ = [
    "__version__",
    "decode",
]



"""
This module contains exceptions for the x.690 protocol
"""


class X690Error(Exception):
    """
    Top-Level exception for everything related to the X690 protocol
    """


class UnexpectedType(X690Error):
    """
    Raised when decoding resulted in an unexpected type.
    """


class IncompleteDecoding(X690Error):
    """
    Raised when decoding did not consume all bytes.

    The junk bytes are stored in the "remainder" attribute
    """

    def __init__(self, message: str, remainder: bytes) -> None:
        super().__init__(message)
        self.remainder = remainder



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['x690']

package_data = \
{'': ['*']}

install_requires = \
['t61codec>=1.0.1,<2.0.0']

extras_require = \
{':python_version < "3.7"': ['dataclasses>=0.7,<0.8'],
 ':python_version < "3.8"': ['importlib-metadata>=3.3.0,<4.0.0']}

setup_kwargs = {
    'name': 'x690',
    'version': '0.5.0a0',
    'description': 'Pure Python X.690 implementation',
    'long_description': 'Pure Python `X.690`_ implementation\n===================================\n\n.. image:: https://github.com/exhuma/x690/workflows/Build%20&%20Publish%20Docs/badge.svg?branch=main\n    :alt: Build & Publish Docs\n\n.. _X.690: https://www.itu.int/rec/recommendation.asp?lang=en&parent=T-REC-X.690-201508-I\n\n\nThis module contains a pure Python implementation of the "x690" standard for\nBER encoding/decoding. Other encodings are currently unsupported but\npull-requests are welcome.\n\n\nSupporting New Types\n--------------------\n\nSome applications may need to support types which are not defined in the X.690\nstandard. This is supported by this library but the types must be defined and\nregistered.\n\nTo register a type, simply subclass ``x690.types.Type``. This will take care of\nthe registration. Make sure that your new type is imported before using it.\n\nNew types should define the following 3 class-variables:\n\n**TYPECLASS**\n    A value from ``x690.util.TypeClass``\n**NATURE**\n    A value from ``x690.util.TypeNature``\n**TAG**\n    A numerical identifier for the type\n\nRefer to the x690 standard for more details on these values. As a general\nrule-of-thumb you can assume that the class is either "context" or\n"application" (it might be good to keep the "universal" class reserved for\nx690). The nature should be "primitive" for simple values and "constructed" for\ncomposed types. The tag is free to choose as long as you don\'t overlap with an\nexisting type.\n\nTo convert raw-bytes into a Python object, override ``x690.Type.decode_raw``\nand conversely also ``x690.Type.encode_raw``. Refer to the docstrings for more\ndetails.\n\n\nReverse Engineering Bytes\n-------------------------\n\nAll types defined in the ``x690`` library provide a ``.pretty()`` method which\nreturns a prettyfied string.\n\nIf you are confronted with a bytes-object encoded using X.690 but don\'t have\nany documentation, you can write the following loop::\n\n    from x690 import decode\n\n    data = open("mydatafile.bin", "rb").read()\n\n    value, nxt = decode(data)\n    print(value.pretty())\n\n    while nxt < len(data):\n        value, nxt = decode(data, nxt)\n        print(value.pretty())\n\nThis should get you started.\n\nIf the data contain non-standard types, they will get detected as "UnknownType"\nand will print out the type-class, nature and tag in the pretty-printed block.\n\nThis will allow you to define your own subclass of ``x690.types.Type`` using\nthose values. Override ``decode(...)`` in that class to handle the unknown\ntype.\n\n\nExamples\n========\n\nEncoding to bytes\n-----------------\n\nEncoding to bytes can be done by simply calling the Python builting ``bytes()``\non instances from ``x690.types``:\n\nEncoding of a single value\n~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n.. code:: python\n\n    >>> import x690.types as t\n    >>> myvalue = t.Integer(12)\n    >>> asbytes = bytes(myvalue)\n    >>> repr(asbytes)\n    b\'\\x02\\x01\\x0c\'\n\nEncoding of a composite value using Sequence\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n.. code:: python\n\n    >>> import x690.types as t\n    >>> myvalue = t.Sequence(\n    ...     t.Integer(12),\n    ...     t.Integer(12),\n    ...     t.Integer(12),\n    ... )\n    >>> asbytes = bytes(myvalue)\n    >>> repr(asbytes)\n    b\'0\\t\\x02\\x01\\x0c\\x02\\x01\\x0c\\x02\\x01\\x0c\'\n\n\nDecoding from bytes\n~~~~~~~~~~~~~~~~~~~\n\nDecode bytes by calling ``x690.types.decode`` on your byte data. This will\nreturn a tuple where the first value contains the decoded object, and the\nsecond one will contain any remaining bytes which were not decoded.\n\n.. code:: python\n\n    >>> import x690\n    >>> data = b\'0\\t\\x02\\x01\\x0c\\x02\\x01\\x0c\\x02\\x01\\x0c\'\n    >>> decoded, nxt = x690.decode(data)\n    >>> decoded\n    Sequence(Integer(12), Integer(12), Integer(12))\n    >>> nxt\n    11\n\n\nType-Hinting & Enforcing\n~~~~~~~~~~~~~~~~~~~~~~~~\n\n**New in 0.3.0**\n\nWhen decoding bytes, it is possible to specify an expcted type which does two\nthings: Firstly, it tells tools like ``mypy`` what the return type will be and\nsecondly, it runs an internal type-check which *ensures* that the returned\nvalue is of the expected type. ``x690.exc.UnexpectedType`` is raised otherwise.\n\nThis does of course only work if you know the type in advance.\n\n.. code:: python\n\n    >>> import x690\n    >>> import x690.types as t\n    >>> data = b\'0\\t\\x02\\x01\\x0c\\x02\\x01\\x0c\\x02\\x01\\x0c\'\n    >>> decoded, nxt = x690.decode(data, enforce_type=t.Sequence)\n    >>> decoded\n    Sequence(Integer(12), Integer(12), Integer(12))\n    >>> nxt\n    11\n\n\nStrict Decoding\n~~~~~~~~~~~~~~~\n\n**New in 0.3.0**\n\nWhen decoding using ``decode`` and you don\'t expect any remaining bytes, use\n``strict=True`` which will raise ``x690.exc.IncompleteDecoding`` if there\'s any\nremaining data.\n\n.. code:: python\n\n    >>> import x690\n    >>> data = b\'0\\t\\x02\\x01\\x0c\\x02\\x01\\x0c\\x02\\x01\\x0cjunk-bytes\'\n    >>> decoded, nxt = x690.decode(data, strict=True)\n    Traceback (most recent call last):\n      ...\n    x690.exc.IncompleteDecoding: Strict decoding still had 10 remaining bytes!\n',
    'author': 'Michel Albert',
    'author_email': 'michel@albert.lu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://exhuma.github.io/x690/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
Utility functions for working with the X.690 and related standards.
"""

from binascii import hexlify, unhexlify
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, NamedTuple

from .exc import X690Error

if TYPE_CHECKING:  # pragma: no cover
    # pylint: disable=unused-import, cyclic-import
    from typing import Dict, List, Tuple, Union

    from .types import Type

#: String to be used for indenting nested items during "pretty()" calls
INDENT_STRING = "  "


class Length(str, Enum):
    """
    A simple "namespace" to avoid magic values for indefinite lengths.
    """

    INDEFINITE = "indefinite"


class TypeClass(str, Enum):
    UNIVERSAL = "universal"
    APPLICATION = "application"
    CONTEXT = "context"
    PRIVATE = "private"


class TypeNature(str, Enum):
    PRIMITIVE = "primitive"
    CONSTRUCTED = "constructed"


class LengthInfo(NamedTuple):
    length: int
    offset: int


class ValueMetaData(NamedTuple):
    bounds: slice
    next_value_index: int


@dataclass
class TypeInfo:
    """
    Decoded structure for an X.690 "type" octet. Example::

        >>> TypeInfo.from_bytes(b'\\x30')
        TypeInfo(cls=<TypeClass.UNIVERSAL: 'universal'>, nature=<TypeNature.CONSTRUCTED: 'constructed'>, tag=16)

    The structure contains 3 fields:

    cls
        The typeclass (a value taken from :py:class:`~.TypeClass`)

    nature
        A value taken from :py:`~.TypeNature`

    tag
        The actual type identifier.

    The instance also keeps the raw value as it was seen in the ``_raw_value``
    attribute.
    """

    cls: TypeClass
    nature: TypeNature
    tag: int

    _raw_value = None

    @staticmethod
    def from_bytes(data):
        # type: ( Union[int, bytes] ) -> TypeInfo
        """
        Given one octet, extract the separate fields and return a TypeInfo
        instance::

            >>> TypeInfo.from_bytes(b'\\x30')
            TypeInfo(cls=<TypeClass.UNIVERSAL: 'universal'>, nature=<TypeNature.CONSTRUCTED: 'constructed'>, tag=16)
        """
        # pylint: disable=attribute-defined-outside-init
        # pylint: disable=protected-access

        if isinstance(data, (bytes, bytearray)):
            data = int.from_bytes(data, "big")
        if data == 0b11111111:
            raise NotImplementedError(
                "Long identifier types are not yet " "implemented"
            )
        cls_hint = (data & 0b11000000) >> 6
        pc_hint = (data & 0b00100000) >> 5
        value = data & 0b00011111

        if cls_hint == 0b00:
            cls = TypeClass.UNIVERSAL
        elif cls_hint == 0b01:
            cls = TypeClass.APPLICATION
        elif cls_hint == 0b10:
            cls = TypeClass.CONTEXT
        elif cls_hint == 0b11:
            cls = TypeClass.PRIVATE
        else:
            pass  # Impossible case (2 bits can only have 4 combinations).

        nature = TypeNature.CONSTRUCTED if pc_hint else TypeNature.PRIMITIVE

        instance = TypeInfo(cls, nature, value)
        instance._raw_value = data
        return instance

    def __bytes__(self):
        # type: () -> bytes
        # pylint: disable=invalid-name
        if self.cls == TypeClass.UNIVERSAL:
            cls = 0b00
        elif self.cls == TypeClass.APPLICATION:
            cls = 0b01
        elif self.cls == TypeClass.CONTEXT:
            cls = 0b10
        elif self.cls == TypeClass.PRIVATE:
            cls = 0b11
        else:
            raise ValueError("Unexpected class for type info")

        if self.nature == TypeNature.CONSTRUCTED:
            nature = 0b01
        elif self.nature == TypeNature.PRIMITIVE:
            nature = 0b00
        else:
            raise ValueError("Unexpected primitive/constructed for type info")

        output = cls << 6 | nature << 5 | self.tag
        return bytes([output])


def encode_length(value):
    # type: (int) -> bytes
    """
    This function encodes the length of a variable into bytes conforming to the
    rules defined in :term:`X.690`: The "length" field must be specially
    encoded for values above 127.  Additionally, from :term:`X.690`:

        8.1.3.2 A sender shall:

            a) use the definite form (see 8.1.3.3) if the encoding is
               primitive;
            b) use either the definite form (see 8.1.3.3) or the indefinite
               form (see 8.1.3.6), a sender's option, if the encoding is
               constructed and all immediately available;
            c) use the indefinite form (see 8.1.3.6) if the encoding is
               constructed and is not all immediately available.

    See also: https://en.wikipedia.org/wiki/X.690#Length_octets

    Example::

        >>> encode_length(16)    # no need for special encoding.
        b'\\x10'
        >>> encode_length(200)   # > 127, needs to be specially encoded.
        b'\\x81\\xc8'
    """
    if value == Length.INDEFINITE:  # type: ignore
        return bytes([0b10000000])

    if value < 127:
        return bytes([value])

    output = []  # type: List[int]
    while value > 0:
        value, remainder = value // 256, value % 256
        output.insert(0, remainder)

    # prefix length information
    output = [0b10000000 | len(output)] + output
    return bytes(output)


def get_value_slice(data: bytes, index: int = 0) -> ValueMetaData:
    """
    Helper method to extract lightweight information about value locations in
    a data-stream.

    The function returns both a slice at which a value can be found, and the
    index at which the next value can be found.
    """
    length, offset = decode_length(data, index + 1)
    if length == -1:
        start = index + 2
        end = data.find(b"\x00\x00", index)
        nex_index = end + 2
    else:
        start = index + 1 + offset
        end = index + 1 + offset + length
        nex_index = end
    value_slice = slice(start, end)
    if end > len(data):
        raise X690Error(
            "Invalid Slice %r (data length=%r)" % (value_slice, len(data))
        )
    return ValueMetaData(value_slice, nex_index)


def decode_length(data, index=0):
    # type: ( bytes, int ) -> LengthInfo
    """
    Given a bytes object, which starts with the length information of a TLV
    value, returns a namedtuple with the length and the number of bytes which
    contained the length information. So, given a TLV value, this function
    takes the "LV" part as input, parses the length information and returns
    the length plus the number of bytes which need to be skipped to arrive at
    the beginning of the value.

    For values which are longer than 127 bytes, the length must be encoded into
    an unknown amount of "length" bytes. This function reads as many bytes as
    needed for the length. Consuming bytes for the value must therefore start
    after the bytes containing the length info. The second value returned
    from this function includes this information.

    The second argument (*index*) tells the function at which position to
    look for the length information.

    Examples::

        >>> # length > 127, consume multiple length bytes
        >>> decode_length(b'\\x81\\xc8...')
        LengthInfo(length=200, offset=2)

        >>> # length <= 127, consume one length byte
        >>> decode_length(b'\\x10...')
        LengthInfo(length=16, offset=1)

    TODO: Upon rereading this, I wonder if it would not make more sense to take
          the complete TLV content as input.
    """
    data0 = data[index]
    if data0 == 0b11111111:
        # reserved
        raise NotImplementedError("This is a reserved case in X690")

    if data0 & 0b10000000 == 0:
        # definite short form
        output = int.from_bytes([data0], "big")
        offset = 1
    elif data0 ^ 0b10000000 == 0:
        # indefinite form
        output = -1
        offset = -1
    else:
        # definite long form
        num_octets = int.from_bytes([data0 ^ 0b10000000], "big")
        value_octets = data[index + 1 : index + num_octets + 1]
        output = int.from_bytes(value_octets, "big")
        offset = num_octets + 1
    return LengthInfo(output, offset)


def visible_octets(data):
    # type: ( bytes ) -> str
    """
    Returns a geek-friendly (hexdump)  output of a bytes object.

    Developer note:
        This is not super performant. But it's not something that's supposed to
        be run during normal operations (mostly for testing and debugging).  So
        performance should not be an issue, and this is less obfuscated than
        existing solutions.

    Example::

        >>> from os import urandom
        >>> print(visible_octets(urandom(40)))  # doctest: +SKIP
        99 1f 56 a9 25 50 f7 9b  95 7e ff 80 16 14 88 c5   ..V.%P...~......
        f3 b4 83 d4 89 b2 34 b4  71 4e 5a 69 aa 9f 1d f8   ......4.qNZi....
        1d 33 f9 8e f1 b9 12 e9                            .3......

    """
    hexed = hexlify(data).decode("ascii")
    tuples = ["".join((a, b)) for a, b in zip(hexed[::2], hexed[1::2])]
    line = []
    output = []
    ascii_column = []
    for idx, octet in enumerate(tuples):
        line.append(octet)
        # only use printable characters in ascii output
        ascii_column.append(octet if 32 <= int(octet, 16) < 127 else "2e")
        if (idx + 1) % 8 == 0:
            line.append("")
        if (idx + 1) % 8 == 0 and (idx + 1) % 16 == 0:
            raw_ascii = unhexlify("".join(ascii_column))
            raw_ascii = raw_ascii.replace(b"\\n z", b".")
            ascii_column = []
            output.append(
                "%-50s %s" % (" ".join(line), raw_ascii.decode("ascii"))
            )
            line = []
    raw_ascii = unhexlify("".join(ascii_column))
    raw_ascii = raw_ascii.replace(b"\\n z", b".")
    output.append("%-50s %s" % (" ".join(line), raw_ascii.decode("ascii")))
    line = []
    return "\n".join(output)


def wrap(text: str, header: str, depth: int) -> str:
    """
    Wraps *text* in a border with a *header* and indented by *indent*.
    """
    prefix = INDENT_STRING * depth
    box_width = max(len(header), max(len(x) for x in text.splitlines()))
    border1 = prefix + "┌─" + ("─" * box_width) + "─┐"
    border2 = prefix + "├─" + ("─" * box_width) + "─┤"
    border3 = prefix + "└─" + ("─" * box_width) + "─┘"
    fmt = "%s│ %%-%ds │" % (prefix, box_width)
    wrapped = [(fmt % line) for line in text.splitlines()]
    output = "\n".join([border1, fmt % header, border2] + wrapped + [border3])
    return output
